# ----------------------------------------------------------------------------
# Copyright (c) 2018-2021, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import os

from qiime2.sdk.usage import Usage, UsageVariable

from q2galaxy.core.util import XMLNode
from q2galaxy.core.templaters.helpers import signature_to_galaxy


def collect_test_data(action, test_dir):
    for idx, example in enumerate(action.examples.values()):
        use = GalaxyTestUsage(example_path=(action, idx), write_dir=test_dir)
        example(use)
        yield from use.created_files


class GalaxyTestUsageVariable(UsageVariable):
    def __init__(self, name, factory, var_type, usage, prefix):
        super().__init__(name, factory, var_type, usage)
        self.prefix = prefix

    def write_file(self, write_dir):
        basename = self.to_interface_name()
        path = os.path.join(write_dir, basename)

        if not os.path.exists(path):
            status = {'status': 'created', 'type': 'file', 'path': path}
        else:
            status = {'status': 'updated', 'type': 'file', 'path': path}

        self.factory().save(path)
        return status

    def to_interface_name(self):
        ext_map = {'artifact': 'qza',
                   'visualization': 'qzv',
                   'metadata': 'tsv'}
        if self.var_type in ext_map:
            ext = ext_map[self.var_type]
            return '.'.join([self.prefix, self.name, ext])
        elif self.var_type == 'column' and hasattr(self, '_q2galaxy_ref'):
            return self._q2galaxy_ref
        else:
            return "UNSUPPORTED: " + self.name

    def assert_output_type(self, semantic_type):
        self._galaxy_has_line_matching(path='metadata.yaml',
                                       expression=f'type: {semantic_type}')

    def assert_has_line_matching(self, path, expression):
        self._galaxy_has_line_matching(path=f'data\\/{path}',
                                       expression=expression)

    def _galaxy_has_line_matching(self, path, expression):
        output = self.use.output_lookup[self.name]

        contents = output.find('assert_contents')
        if contents is None:
            contents = XMLNode('assert_contents')
            output.append(contents)

        path = (r'[0-9a-f]{8}-[0-9a-f]{4}-[4][0-9a-f]{3}-[89ab][0-9a-f]'
                r'{3}-[0-9a-f]{12}\/') + path
        archive = contents.find(f'has_archive_member[@path="{path}"]')
        if archive is None:
            archive = XMLNode('has_archive_member', path=path)
            contents.append(archive)

        archive.append(XMLNode('has_line_matching', expression=expression))


class GalaxyTestUsage(Usage):
    def __init__(self, example_path, write_dir=None):
        super().__init__()
        self.prefix = f'{example_path[0].id}.test{example_path[1]}'
        self.xml = XMLNode('test')
        self.output_lookup = {}
        self.write_dir = write_dir
        self.created_files = []

    def usage_variable(self, name, factory, var_type):
        return GalaxyTestUsageVariable(name, factory, var_type, self,
                                       self.prefix)

    def init_artifact(self, name, factory):
        var = super().init_artifact(name, factory)

        if self.write_dir is not None:
            status = var.write_file(self.write_dir)
            self.created_files.append(status)

        return var

    def init_metadata(self, name, factory):
        var = super().init_metadata(name, factory)

        if self.write_dir is not None:
            status = var.write_file(self.write_dir)
            self.created_files.append(status)

        return var

    def get_metadata_column(self, name, column_name, variable):
        var = super().get_metadata_column(name, column_name, variable)

        md = variable.execute()
        # 1-based index, first col is IDs, second col is the first data column
        col_idx = str(list(md.columns.keys()).index(column_name) + 2)
        var._q2galaxy_ref = (variable.to_interface_name(), col_idx)

        return var

    def merge_metadata(self, name, *variables):
        var = super().merge_metadata(name, *variables)
        # TODO: do /something/ when this is supported.
        return var

    def action(self, action, inputs, outputs):
        vars_ = super().action(action, inputs, outputs)

        sig = action.get_action().signature
        mapped = inputs.map_variables(lambda v: v.to_interface_name())
        for case in signature_to_galaxy(sig, mapped):
            test_xml = case.tests_xml()
            if test_xml is None:
                continue
            if type(test_xml) is not list:
                test_xml = [test_xml]
            for xml in test_xml:
                self.xml.append(xml)

        for output_name, output in outputs.items():
            xml_out = XMLNode('output', name=output_name, ftype='qza')
            self.output_lookup[output] = xml_out
            self.xml.append(xml_out)

        return vars_
